//  Program:  classlist
//  File:     classlist_itr.cpp
//  Author:   Chuck Stewart
//  Purpose:  Build and maintain a list of students enrolled in a
//     class and a waiting list.  This is the list version.

#include <algorithm>
#include <fstream>
#include <iostream>
#include <string>
#include <list>

void enroll_student( const std::string& id_number, 
                     unsigned int max_students,
                     std::list<std::string>& enrolled,
                     std::list<std::string>& waiting );
void remove_student( const std::string& id_number, 
                     unsigned int max_students,
                     std::list<std::string>& enrolled,
                     std::list<std::string>& waiting );

int
main( int argc, char* argv[] )
{
  if ( argc != 2 )
    {
      std::cerr << "Usage: " << argv[0] << " enroll.txt\n";
      return 1;
    }

  std::ifstream enroll_str( argv[1] );
  if ( !enroll_str )
    {
      std::cerr << "Failed to open " << argv[1] << '\n';
      return 1;
    }

  // Read in the maximum number of students in the course
  unsigned int max_students;
  enroll_str >> max_students;
  
  // Initialize the list
  std::list<std::string> enrolled;
  std::list<std::string> waiting;

  //  The input file is very simple.  Each line contains a 0 or a 1,
  //  followed by a student id.  A 0 indicates that a student wants to
  //  be enrolled in the course and a 1 indicates that a student wants
  //  to be removed.   We will assume that the data are correct.
  //
  //  Here is the invariant about the state of the enrolled and
  //  waiting lists:
  //
  //   (1) enrolled contains the students already in the course,
  //   (2) waiting contains students who would like to be admitted
  //       into the course when the course is full;  they are admitted
  //       to the course in the order they requested admission,
  //   (3) enrolled.size() <= max_students,
  //   (4) if the course is not filled ( enrolled.size() !=
  //       max_students ) then waiting is empty

  int option;
  std::string id;
  while ( enroll_str >> option >> id )
    {
      if ( option == 0 )
	enroll_student( id, max_students, enrolled, waiting );
      else
	remove_student( id, max_students, enrolled, waiting ); 
    }

  enrolled.sort( );
  std::cout << "\nAt the end of the enrollment period, the following students\n"
	    << "are in the class:\n\n";
  for ( std::list<std::string>::iterator p = enrolled.begin();
	p != enrolled.end(); ++p )
    std::cout << *p << '\n';
  //  for ( unsigned int i=0; i<enrolled.size(); ++i )
  //    std::cout << enrolled[i] << '\n';

  if ( ! waiting.empty() )
    {
      std::cout << "\nStudents are on the waiting list in the following order:\n";
      for ( std::list<std::string>::iterator p = waiting.begin();
	    p != waiting.end(); ++p )
	std::cout << *p << std::endl;
    }
  return 0;
}


// Enroll a student in the course if there is room and if the student
// is not already in the course or on the waiting list.

void
enroll_student( const std::string& id_number, 
                unsigned int max_students,
                std::list<std::string>& enrolled,
                std::list<std::string>& waiting )
{
  // Check to see if the student is already enrolled.  If so, output a
  // message and return.

  std::list<std::string>::iterator p, q;
  for ( p = enrolled.begin(); p!=enrolled.end(); ++p )
    {
      if ( *p == id_number )
        {
	  std::cout << "Student " << id_number
		    << " is already enrolled." << std::endl;
          return;
        }
    }

  //  If the course isn't full, add the student.  Then return.

  if ( enrolled.size() < max_students )
    {
      enrolled.push_back( id_number );
      std::cout << "Student " << id_number << " added.  "
		<< enrolled.size() << " students are now in the course."
		<< std::endl;
      return;
    }

  //  Otherwise, we have to deal with the waiting list.  Check to see
  //  if the student is already on the waiting list.

  for ( q = waiting.begin(); q!=waiting.end(); ++q )
    {
      if ( *q == id_number )
        {
	  std::cout << "Student " << id_number
		    << " is already on the waiting list." << std::endl;
          return;
        }
    }

  //  Otherwise, add the student to the waiting list.

  waiting.push_back( id_number );
  std::cout << "The course is full.  Student " << id_number 
	    << " is in position " << waiting.size() << " on the waiting list."
	    << std::endl;
}


//  Remove a student from the course or from the waiting list.  If
//  removing the student from the course opens up a slot, then the
//  first person on the waiting list is placed in the course.

void 
remove_student( const std::string& id_number, 
                unsigned int max_students,
                std::list<std::string>& enrolled,
                std::list<std::string>& waiting )
{
  //  Check to see if the student is on the course list.
  bool found = false;
  std::list<std::string>::iterator loc = enrolled.begin();
  while ( !found && loc != enrolled.end() )
    {
      found = (*loc == id_number);
      if ( !found ) ++ loc;
    }

  //  If so, remove the student and see if a student can be taken from
  //  the waiting list.
  if ( found )
    {
      enrolled.erase( loc );
      std::cout << "Student " << id_number << " removed from the course.  ";
      if ( waiting.size() > 0 )
        {
          enrolled.push_back( waiting.front() );
	  std::cout << "Student " << waiting.front()
		    << " added from the waiting list, which now contains ";
	  waiting.erase( waiting.begin() );
	  std::cout << waiting.size() << " students." << std::endl;
        }
      else
        {
	  std::cout << enrolled.size() << " students are now in the course."
		    << std::endl;
        }
    }

  //  If not, check to see if the student is on the waiting list
  else 
    {
      found = false;
      loc = waiting.begin();
      while ( !found && loc != waiting.end() )
        {
          found = ( *loc == id_number );
          if ( ! found ) ++ loc;
        }

      if ( found )
        {
	  waiting.erase( loc );
	  std::cout << "Student " << id_number
		    << " removed from the waiting list.\n"
		    << waiting.size() << " students remain on the waiting list."
		    << std::endl;
        }
      else
        {
	  std::cout << "Student " << id_number
		    << " is in neither the course nor on the waiting list"
		    << std::endl;
        }
    }
}
        

      
      
